from pathlib import Path
import csv
import inspect
import json
from io import StringIO
import pathlib
import warnings
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial import Delaunay
from photompy import read_ies_data, plot_ies, total_optical_power
from .trigonometry import to_cartesian, to_polar, attitude
from ._helpers import NumpyEncoder, parse_json


class Lamp:
    """
    Represents a lamp with properties defined by a photometric data file.
    This class handles the loading of IES photometric data, orienting the lamp in 3D space,
    and provides methods for moving, rotating, and aiming the lamp

    Parameters:
    lamp_id: str
        A unique identifier for the lamp. Only required parameter.
    name: str
        Non-unique display name for the lamp
    filename: Path, str, or None
        If None or not pathlike, `filedata` must not be None
    filedata: Path or bytes or None
        Set by `filename` if filename is pathlike.
    x, y, z: floats
        Sets initial position of lamp in cartesian space
    angle: float
        Sets lamps initial rotation on its own axis.
    aimx, aimy, aimz: floats
        Sets initial aim point of lamp in cartesian space.
    spectra_source: Path or bytes or None
        Data source for spectra
    spectra_weight_source: Path or bytes or None
        Data source for spectral weighting
    spectra: dict
        Dictionary where keys are labels and values are arraylikes of shape (2,N) where N = the number of
        (wavelength, relative intensity) pairs that define the lamp's spectra. Set by `spectra_source` if provided,
        otherwise None, or may be passed directly. An unweighted spectra will have the key 'Unweighted'. If an unlabeled
        arraylike is passed, it will be assumed to be unweighted and weighted spectra will be calculated on that basis, if
        a `spectral_weight_source` was passed.
    spectral_weightings: dict
        Dictionary where keys are labels for a particular spectral weighting, and values are arraylikes of shape (2,N)
        where N = the number of (wavelength, relative intensity) pairs.
    intensity_units: str
        generally assumed to be `mW/Sr`. Future features will support other units, like uW/cm2
    radiation_type: str
        set from ies file keywords. Currently, only UVC222 is supported for GUV features.
    enabled: bool
        determines if lamp participates in calculations
    """

    def __init__(
        self,
        lamp_id,
        name=None,
        filename=None,
        filedata=None,
        x=None,
        y=None,
        z=None,
        angle=None,
        aimx=None,
        aimy=None,
        aimz=None,
        spectra_source=None,
        spectra=None,
        spectral_weight_source=None,
        spectral_weightings=None,
        intensity_units=None,
        radiation_type=None,
        max_irradiances=None,
        enabled=None,
    ):
        self.lamp_id = lamp_id
        self.name = lamp_id if name is None else name
        self.enabled = True if enabled is None else enabled
        # position
        self.x = 0.0 if x is None else x
        self.y = 0.0 if y is None else y
        self.z = 0.0 if z is None else z
        self.position = np.array([self.x, self.y, self.z])
        # orientation
        self.angle = 0.0 if angle is None else angle
        self.aimx = self.x if aimx is None else aimx
        self.aimy = self.y if aimy is None else aimy
        self.aimz = self.z - 1.0 if aimz is None else aimz
        self.aim(self.aimx, self.aimy, self.aimz)  # updates heading and bank

        # misc
        self.intensity_units = "mW/Sr" if intensity_units is None else intensity_units
        self.radiation_type = radiation_type
        # calc zone values will be stored here
        self.max_irradiances = {} if max_irradiances is None else max_irradiances

        # spectral weightings
        self.spectral_weight_source = spectral_weight_source
        self.spectral_weightings = (
            {} if spectral_weightings is None else spectral_weightings
        )
        if (
            self.spectral_weight_source is not None
            and len(self.spectral_weightings) == 0
        ):
            self._load_spectral_weightings()
        # spectra - unweighted and weighted
        self.spectra = {} if spectra is None else spectra
        self.spectra_source = spectra_source
        if self.spectra_source is not None and len(self.spectra) == 0:
            self._load_spectra()
            self._update_spectra()

        # load file and coordinates
        self.filename = filename
        self.filedata = filedata
        self._check_filename()

        # filename is just a label, filedata controls everything.
        if self.filedata is not None:
            self._load()
            self._orient()

    def _load_csv(self, datasource):
        """load csv data from either path or bytes"""
        if isinstance(datasource, (str, pathlib.PosixPath)):
            filepath = Path(datasource)
            filetype = filepath.suffix.lower()
            if filetype != ".csv":
                raise TypeError("Currently, only .csv files are supported")
            csv_data = open(datasource, mode="r", newline="")
        elif isinstance(datasource, bytes):
            # Convert bytes to a string using StringIO to simulate a file
            csv_data = StringIO(datasource.decode("utf-8"))
        else:
            raise TypeError(f"File type {type(datasource)} not valid")
        return csv_data

    def _load_spectra(self):
        """load spectral data from source"""
        csv_data = self._load_csv(self.spectra_source)
        reader = csv.reader(csv_data, delimiter=",")
        # read each line
        spectra = []
        for i, row in enumerate(reader):
            try:
                wavelength, intensity = map(float, row)
                spectra.append((wavelength, intensity))
            except ValueError:
                if i == 0:  # probably a header
                    continue
                else:
                    warnings.warn(f"Skipping invalid datarow: {row}")
        self.spectra["Unweighted"] = np.array(spectra).T

    def load_spectra(self, spectra_source):
        """
        external method to set self.spectra_source and invoke internal method
        _load_spectra to read the source into self.spectra.
        If weightings are present, update weighted spectra also
        If spectra_source is none, self.spectra will reset to empty
        """
        self.spectra_source = spectra_source
        if self.spectra_source is None:
            self.spectra = {}  # reset to empty
        else:
            self._load_spectra()
            if self.spectral_weight_source is not None:
                self._load_spectral_weightings()
                self._update_spectra()

    def _load_spectral_weightings(self):
        """load weightings"""
        csv_data = self._load_csv(self.spectral_weight_source)
        reader = csv.reader(csv_data, delimiter=",")
        headers = next(reader, None)  # get headers

        data = {}
        for header in headers:
            data[header] = []
        for row in reader:
            for header, value in zip(headers, row):
                data[header].append(float(value))

        # update self.spectra
        wavelengths = data[headers[0]]
        for i, (key, val) in enumerate(data.items()):
            if i == 0:
                continue
            else:
                values = np.array(val)
                self.spectral_weightings[key] = np.stack((wavelengths, values)).tolist()

    def _update_spectra(self):
        """
        weight the unweighted spectra by all potential spectral weightings and add to
        the self.spectra dict
        """

        if self.spectra is not None:
            wavelengths = self.spectra["Unweighted"][0]
            intensities = self.spectra["Unweighted"][1]
            maxval = max(intensities)
            for key, val in self.spectral_weightings.items():
                # update weights to match the spectral wavelengths we've got
                weights = np.interp(wavelengths, val[0], val[1])
                # self.spectral_weightings[key] = np.stack((wavelengths, weights))
                # weight spectra
                weighted_intensity = intensities * weights
                ratio = maxval / max(weighted_intensity)
                self.spectra[key] = np.stack(
                    (wavelengths, weighted_intensity * ratio)
                ).tolist()

    def load_weighted_spectra(self, spectral_weight_source):
        """
        external method to set self.spectral_weight_source and invoke internal method
        _load_weighted_spectra to read the weightings and update self.spectra
        """
        self.spectral_weight_source = spectral_weight_source
        self._load_spectral_weightings()
        self._update_spectra()

    def _check_filename(self):
        """
        determine datasource
        if filename is a path AND file exists AND filedata is None, it replaces filedata
        otherwise filedata stays the same
        """
        FILE_IS_PATH = False
        # if filename is string, check if it's a path
        if isinstance(self.filename, (str, pathlib.PosixPath)):
            if Path(self.filename).is_file():
                FILE_IS_PATH = True
        # if filename is a path and exists, it will replace filedata, but only if filedata wasn't specified to begin with
        if FILE_IS_PATH and self.filedata is None:
            self.filedata = self.filename

    def _load(self):
        """
        Loads lamp data from an IES file and initializes photometric properties.
        """
        self.lampdict = read_ies_data(self.filedata)
        self.valdict = self.lampdict["full_vals"]
        self.thetas = self.valdict["thetas"]
        self.phis = self.valdict["phis"]
        self.values = self.valdict["values"]
        self.interpdict = self.lampdict["interp_vals"]

        units_type = self.lampdict["units_type"]
        if units_type == 1:
            self.units = "feet"
        elif units_type == 2:
            self.units = "meters"
        else:
            msg = "Lamp dimension units could not be determined. Your ies file may be malformed. Units of meters are being assumed."
            warnings.warn(msg)
            self.units = "meters"

        self.dimensions = [
            self.lampdict["width"],
            self.lampdict["length"],
            self.lampdict["height"],
        ]
        self.input_watts = self.lampdict["input_watts"]
        self.keywords = self.lampdict["keywords"]
        if "_RADIATIONTYPE" in self.keywords.keys():
            self.radiation_type = self.keywords["_RADIATIONTYPE"]

    def _orient(self):
        """
        Initializes the orientation of the lamp based on its photometric data.
        """

        # true value coordinates
        tgrid, pgrid = np.meshgrid(self.thetas, self.phis)
        tflat, pflat = tgrid.flatten(), pgrid.flatten()
        tflat = 180 - tflat  # to account for reversed z direction
        x, y, z = to_cartesian(tflat, pflat, 1)
        self.coords = np.array([x, y, z]).T

        # photometric web coordinates
        xp, yp, zp = to_cartesian(tflat, pflat, self.values.flatten())
        self.photometric_coords = np.array([xp, yp, zp]).T

    def _recalculate_aim_point(self, dimensions=None, distance=None):
        """
        internal method to call if setting tilt/bank or orientation/heading
        if `dimensions` is passed, `distance` is not used
        """
        distance = 1 if distance is None else distance
        heading_rad = np.radians(self.heading)
        # Correcting bank angle for the pi shift
        bank_rad = np.radians(self.bank - 180)

        # Convert from spherical to Cartesian coordinates
        dx = np.sin(bank_rad) * np.cos(heading_rad)
        dy = np.sin(bank_rad) * np.sin(heading_rad)
        dz = np.cos(bank_rad)
        if dimensions is not None:
            distances = []
            dimx, dimy, dimz = dimensions
            if dx != 0:
                distances.append((dimx - self.x) / dx if dx > 0 else self.x / -dx)
            if dy != 0:
                distances.append((dimy - self.y) / dy if dy > 0 else self.y / -dy)
            if dz != 0:
                distances.append((dimz - self.z) / dz if dz > 0 else self.z / -dz)
            distance = min([d for d in distances])
        self.aim_point = self.position + np.array([dx, dy, dz]) * distance
        self.aimx, self.aimy, self.aimz = self.aim_point

    def get_total_power(self):
        """return the lamp's total optical power"""
        self.total_optical_power = total_optical_power(self.interpdict)
        return self.total_optical_power

    def reload(self, filename=None, filedata=None):
        """replace the ies file without erasing any position/rotation/eing information"""

        self.filename = filename
        self.filedata = filedata
        # if filename is a path, filedata is filename
        self._check_filename

        if self.filedata is not None:
            self._load()
            self._orient()
        else:
            self.lampdict = None
            self.valdict = None
            self.thetas = None
            self.phis = None
            self.values = None
            self.interpdict = None
            self.units = None
            self.dimensions = None
            self.input_watts = None
            self.keywords = None
            self.coords = None
            self.photometric_coords = None
            self.spectra = {}

    def transform(self, coords, scale=1):
        """
        Transforms the given coordinates based on the lamp's orientation and position.
        Applies rotation, then aiming, then scaling, then translation.
        Scale parameter should generally only be used for photometric_coords
        """
        # in case user has updated x y and z
        coords = np.array(attitude(coords.T, roll=0, pitch=0, yaw=self.angle)).T
        coords = np.array(
            attitude(coords.T, roll=0, pitch=self.bank, yaw=self.heading)
        ).T
        coords = (coords.T / scale).T + self.position
        return coords

    def get_cartesian(self, scale=1, sigfigs=9):
        """Return lamp's true position coordinates in cartesian space"""
        return self.transform(self.coords, scale=scale).round(sigfigs)

    def get_polar(self, sigfigs=9):
        """Return lamp's true position coordinates in polar space"""
        cartesian = self.transform(self.coords) - self.position
        return np.array(to_polar(*cartesian.T)).round(sigfigs)

    def move(self, x=None, y=None, z=None):
        """Designate lamp position in cartesian space"""
        # determine new position   selected_lamp.
        x = self.x if x is None else x
        y = self.y if y is None else y
        z = self.z if z is None else z
        position = np.array([x, y, z])
        # update aim point based on new position
        diff = position - self.position
        self.aim_point += diff
        self.aimx, self.aimy, self.aimz = self.aim_point
        # update position
        self.position = position
        self.x, self.y, self.z = self.position
        return self

    def rotate(self, angle):
        """designate lamp orientation with respect to its z axis"""
        self.angle = angle
        return self

    def set_orientation(self, orientation, dimensions=None, distance=None):
        """
        set orientation/heading.
        alternative to setting aim point with `aim`
        distinct from rotation; applies to a tilted lamp. to rotate a lamp along its axis,
        use the `rotate` method
        """
        # orientation = (orientation + 360) % 360
        self.heading = orientation
        self._recalculate_aim_point(dimensions=dimensions, distance=distance)

    def set_tilt(self, tilt, dimensions=None, distance=None):
        """
        set tilt/bank
        alternative to setting aim point with `aim`
        """
        # tilt = (tilt + 360) % 360
        self.bank = tilt
        self._recalculate_aim_point(dimensions=dimensions, distance=distance)

    def aim(self, x=None, y=None, z=None):
        """aim lamp at a point in cartesian space"""
        x = self.aimx if x is None else x
        y = self.aimy if y is None else y
        z = self.aimz if z is None else z
        self.aim_point = np.array([x, y, z])
        self.aimx, self.aimy, self.aimz = self.aim_point
        xr, yr, zr = self.aim_point - self.position
        self.heading = np.degrees(np.arctan2(yr, xr))
        self.bank = np.degrees(np.arctan2(np.sqrt(xr ** 2 + yr ** 2), zr) - np.pi)
        # self.heading = (heading+360)%360
        # self.bank = (bank+360)%360
        return self

    def plot_ies(self, title=""):
        """standard polar plot of an ies file"""
        fig, ax = plot_ies(fdata=self.valdict, title=title)
        return fig, ax

    def plot_spectra(self, title=None, fig=None, figsize=(6.4, 4.8), yscale="linear"):
        """
        plot the spectra of the lamp. at minimum, the unweighted spectra, possibly all
        weighted spectra as well.

        `yscale` is generally either "linear" or "log", but any matplotlib scale is permitted
        """

        if fig is None:
            fig, ax = plt.subplots()
        else:
            ax = fig.axes[0]

        if len(self.spectra) > 0:
            for key, val in self.spectra.items():
                linestyle = "-" if key == "Unweighted" else "--"
                ax.plot(val[0], val[1], label=key, linestyle=linestyle)
            ax.legend()
            ax.grid(True, which="both", ls="--", c="gray", alpha=0.3)
            ax.set_xlabel("Wavelength [nm]")
            ax.set_ylabel("Relative intensity [%]")
            ax.set_yscale(yscale)

            title = self.name if title is None else title
            ax.set_title(title)
        return fig

    def plot_3d(
        self,
        elev=45,
        azim=-45,
        title="",
        figsize=(6, 4),
        show_cbar=False,
        alpha=0.7,
        cmap="rainbow",
        fig=None,
        ax=None,
    ):
        """plot in cartesian 3d space of the true positions of the irradiance values"""
        x, y, z = self.transform(self.coords).T
        intensity = self.values.flatten()
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, projection="3d")
        ax.scatter(x, y, z, c=intensity, cmap="rainbow", alpha=alpha)
        if self.aim_point is not None:
            ax.plot(
                *np.array((self.aim_point, self.position)).T,
                linestyle="--",
                color="black",
                alpha=0.7,
            )
        ax.set_title(title)
        ax.view_init(azim=azim, elev=elev)
        ax.set_xlabel("x")
        ax.set_ylabel("y")
        ax.set_zlabel("z")
        return fig, ax

    def plot_web(
        self,
        elev=30,
        azim=-60,
        title="",
        figsize=(6, 4),
        color="#cc61ff",
        alpha=0.4,
        xlim=None,
        ylim=None,
        zlim=None,
    ):
        """plot photometric web, where distance r is set by the irradiance value"""
        scale = self.values.max()
        x, y, z = self.transform(self.photometric_coords, scale=scale).T
        Theta, Phi, R = to_polar(*self.photometric_coords.T)
        tri = Delaunay(np.column_stack((Theta.flatten(), Phi.flatten())))
        fig = plt.figure(figsize=figsize)
        ax = fig.add_subplot(111, projection="3d")
        ax.plot_trisurf(x, y, z, triangles=tri.simplices, color=color, alpha=alpha)
        if self.aim_point is not None:
            ax.plot(
                *np.array((self.aim_point, self.position)).T,
                linestyle="--",
                color="black",
                alpha=0.7,
            )
        ax.set_title(title)
        if xlim is not None:
            ax.set_xlim(xlim)
        if ylim is not None:
            ax.set_ylim(ylim)
        if zlim is not None:
            ax.set_zlim(zlim)
        ax.view_init(azim=azim, elev=elev)
        return fig, ax

    @classmethod
    def from_json(cls, jsondata):
        keys = list(inspect.signature(cls.__init__).parameters.keys())[1:]
        data = parse_json(jsondata)
        return cls(**{k: v for k, v in data.items() if k in keys})

    def to_json(self):
        # Create a dictionary of all instance variables
        data = {attr: getattr(self, attr) for attr in vars(self)}
        return json.dumps(data, cls=NumpyEncoder)



import inspect
import json
import numpy as np
import matplotlib.pyplot as plt
from photompy import get_intensity_vectorized
from .trigonometry import attitude, to_polar
from ._helpers import NumpyEncoder, parse_json


class CalcZone(object):
    """
    Base class representing a calculation zone.

    This class provides a template for setting up zones within which various
    calculations related to lighting conditions are performed. Subclasses should
    provide specific implementations of the coordinate setting method.

    NOTE: I changed this from an abstract base class to an object superclass
    to make it more convenient to work with the website, but this class doesn't really
    work on its own

    Parameters:
    --------
    zone_id: str
        identification tag for internal tracking
    name: str, default=None
        externally visible name for zone
    dimensions: array of floats, default=None
        array of len 2 if CalcPlane, of len 3 if CalcVol
    offset: bool, default=True
    fov80: bool, default=False
        apply 80 degree field of view filtering - used for calculating eye limits
    vert: bool, default=False
        calculate vertical irradiance only
    horiz: bool, default=False
        calculate horizontal irradiance only
    dose: bool, default=False
        whether to calculate a dose over N hours or just fluence
    hours: float, default = 8.0
        number of hours to calculate dose over. Only relevant if dose is True.
    enabled: bool, default = True
        whether or not the calc zone is enabled for calculations
    """

    def __init__(
        self,
        zone_id,
        name=None,
        offset=None,
        fov80=None,
        vert=None,
        horiz=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
    ):
        self.zone_id = zone_id
        self.name = zone_id if name is None else name
        self.offset = True if offset is None else offset
        self.fov80 = False if fov80 is None else fov80
        self.vert = False if vert is None else vert
        self.horiz = False if horiz is None else horiz
        self.dose = False if dose is None else dose
        if self.dose:
            self.units = "mJ/cm2"
        else:
            self.units = "uW/cm2"
        self.hours = 8.0 if hours is None else hours  # only used if dose is true
        self.enabled = True if enabled is None else enabled
        self.show_values = True if show_values is None else show_values

        # these will all be calculated after spacing is set, which is set in the subclass
        self.calctype = "Zone"
        self.x1 = None
        self.x2 = None
        self.y1 = None
        self.y2 = None
        self.z1 = None
        self.z2 = None
        self.height = None
        self.spacing = None
        self.x_spacing = None
        self.y_spacing = None
        self.z_spacing = None
        self.num_points = None
        self.xp = None
        self.yp = None
        self.zp = None
        self.coords = None
        self.values = None

    @classmethod
    def from_json(cls, jsondata):
        keys = list(inspect.signature(cls.__init__).parameters.keys())[1:]
        data = parse_json(jsondata)
        return cls(**{k: v for k, v in data.items() if k in keys})

    def to_json(self):
        # Create a dictionary of all instance variables
        data = {attr: getattr(self, attr) for attr in vars(self)}
        return json.dumps(data, cls=NumpyEncoder)

    def set_dimensions(self, dimensions):
        raise NotImplementedError

    def set_spacing(self, spacing):
        raise NotImplementedError

    def set_offset(self, offset):
        if type(offset) is not bool:
            raise TypeError("Offset must be either True or False")
        self.offset = offset
        self._update

    def set_value_type(self, dose):
        """
        if true values will be in dose over time
        if false
        """
        if type(dose) is not bool:
            raise TypeError("Dose must be either True or False")

        # convert values if they need converting
        if self.values is not None:
            if dose and not self.dose:
                self.values = self.values * 3.6 * self.hours
            elif self.dose and not dose:
                self.values = self.values / (3.6 * self.hours)

        self.dose = dose
        if self.dose:
            self.units = "mJ/cm2"
        else:
            self.units = "uW/cm2"

    def set_dose_time(self, hours):
        if type(hours) not in [float, int]:
            raise TypeError("Hours must be numeric")
        self.hours = hours

    def calculate_values(self, lamps: list):
        """
        Calculate and return irradiance values at all coordinate points within the zone.
        """
        total_values = np.zeros(self.coords.shape[0])
        for lamp_id, lamp in lamps.items():
            if lamp.filedata is not None and lamp.enabled:
                # determine lamp placement + calculate relative coordinates
                rel_coords = self.coords - lamp.position
                # store the theta and phi data based on this orientation
                Theta0, Phi0, R0 = to_polar(*rel_coords.T)
                # apply all transformations that have been applied to this lamp, but in reverse
                rel_coords = np.array(
                    attitude(rel_coords.T, roll=0, pitch=0, yaw=-lamp.heading)
                ).T
                rel_coords = np.array(
                    attitude(rel_coords.T, roll=0, pitch=-lamp.bank, yaw=0)
                ).T
                rel_coords = np.array(
                    attitude(rel_coords.T, roll=0, pitch=0, yaw=-lamp.angle)
                ).T
                Theta, Phi, R = to_polar(*rel_coords.T)
                values = get_intensity_vectorized(Theta, Phi, lamp.interpdict) / R ** 2

                if self.fov80:
                    values[Theta0 < 50] = 0
                if self.vert:
                    values *= np.sin(np.radians(Theta0))
                if self.horiz:
                    values *= np.cos(np.radians(Theta0))
                if lamp.intensity_units == "mW/Sr":
                    total_values += values / 10  # convert from mW/Sr to uW/cm2
                else:
                    raise KeyError("Units not recognized")

                # save the max value to the lamp object
                lamp.max_irradiances[self.zone_id] = total_values.max()

        # I have no earthly idea why it is necessary to do it this way
        if len(self.num_points) == 3:
            total_values = total_values.reshape(
                (self.num_points[1], self.num_points[0], self.num_points[2])
            )
        elif len(self.num_points) == 2:
            total_values = total_values.reshape(
                (self.num_points[1], self.num_points[0])
            )
        total_values = np.ma.masked_invalid(
            total_values
        )  # mask any nans near light source
        self.values = total_values
        # convert to dose
        if self.dose:
            self.values = self.values * 3.6 * self.hours

        return self.values


class CalcVol(CalcZone):
    """
    Represents a volumetric calculation zone.
    A subclass of CalcZone designed for three-dimensional volumetric calculations.
    """

    def __init__(
        self,
        zone_id,
        name=None,
        x1=None,
        x2=None,
        y1=None,
        y2=None,
        z1=None,
        z2=None,
        x_spacing=None,
        y_spacing=None,
        z_spacing=None,
        offset=None,
        fov80=None,
        vert=None,
        horiz=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
    ):

        super().__init__(
            zone_id=zone_id,
            name=name,
            offset=offset,
            fov80=fov80,
            vert=vert,
            horiz=horiz,
            dose=dose,
            hours=hours,
            enabled=enabled,
            show_values=show_values,
        )
        self.calctype = "Volume"
        self.x1 = 0 if x1 is None else x1
        self.x2 = 6 if x2 is None else x2
        self.y1 = 0 if y1 is None else y1
        self.y2 = 4 if y2 is None else y2
        self.z1 = 0 if z1 is None else z1
        self.z2 = 2.7 if z2 is None else z2
        self.x_spacing = 0.1 if x_spacing is None else x_spacing
        self.y_spacing = 0.1 if y_spacing is None else y_spacing
        self.z_spacing = 0.1 if z_spacing is None else z_spacing
        self._update()

    def set_dimensions(self, x1=None, x2=None, y1=None, y2=None, z1=None, z2=None):
        self.x1 = self.x1 if x1 is None else x1
        self.x2 = self.x2 if x2 is None else x2
        self.y1 = self.y1 if y1 is None else y1
        self.y2 = self.y2 if y2 is None else y2
        self.z1 = self.z1 if z1 is None else z1
        self.z2 = self.z2 if z2 is None else z2
        self._update()

    def set_spacing(self, x_spacing=None, y_spacing=None, z_spacing=None):
        self.x_spacing = self.x_spacing if x_spacing is None else x_spacing
        self.y_spacing = self.y_spacing if y_spacing is None else y_spacing
        self.z_spacing = self.z_spacing if z_spacing is None else z_spacing
        self._update()

    def _update(self):
        """
        Update the number of points based on the spacing, and then the points
        """
        numx = int((self.x2 - self.x1) / self.x_spacing)
        numy = int((self.y2 - self.y1) / self.y_spacing)
        numz = int((self.z2 - self.z1) / self.z_spacing)
        self.num_points = np.array([numx, numy, numz])
        if self.offset:
            xpoints = np.linspace(
                self.x1 + (self.x_spacing / 2), self.x2 - (self.x_spacing / 2), numx
            )
            ypoints = np.linspace(
                self.y1 + (self.y_spacing / 2), self.y2 - (self.y_spacing / 2), numy
            )
            zpoints = np.linspace(
                self.z1 + (self.z_spacing / 2), self.z2 - (self.z_spacing / 2), numz
            )
        else:
            xpoints = np.linspace(self.x1, self.x2, numx)
            ypoints = np.linspace(self.y1, self.y2, numy)
            zpoints = np.linspace(self.z1, self.z2, numz)
        self.points = [xpoints, ypoints, zpoints]
        self.xp, self.yp, self.zp = self.points
        X, Y, Z = [grid.reshape(-1) for grid in np.meshgrid(*self.points)]
        self.coords = np.array((X, Y, Z)).T


class CalcPlane(CalcZone):
    """
    Represents a planar calculation zone.
    A subclass of CalcZone designed for two-dimensional planar calculations at a specific height.
    """

    def __init__(
        self,
        zone_id,
        name=None,
        x1=None,
        x2=None,
        y1=None,
        y2=None,
        height=None,
        x_spacing=None,
        y_spacing=None,
        offset=None,
        fov80=None,
        vert=None,
        horiz=None,
        dose=None,
        hours=None,
        enabled=None,
        show_values=None,
    ):

        super().__init__(
            zone_id=zone_id,
            name=name,
            offset=offset,
            fov80=fov80,
            vert=vert,
            horiz=horiz,
            dose=dose,
            hours=hours,
            enabled=enabled,
            show_values=show_values,
        )
        self.calctype = "Plane"
        self.height = 1.9 if height is None else height
        self.x1 = 0 if x1 is None else x1
        self.x2 = 6 if x2 is None else x2
        self.y1 = 0 if y1 is None else y1
        self.y2 = 4 if y2 is None else y2
        self.x_spacing = 0.1 if x_spacing is None else x_spacing
        self.y_spacing = 0.1 if y_spacing is None else y_spacing
        self._update()

    def set_height(self, height):
        """set height of calculation plane. currently we only support vertical planes"""
        if type(height) not in [float, int]:
            raise TypeError("Height must be numeric")
        self.height = height
        self._update()

    def set_dimensions(self, x1=None, x2=None, y1=None, y2=None):
        """set the dimensions and update the coordinate points"""
        self.x1 = self.x1 if x1 is None else x1
        self.x2 = self.x2 if x2 is None else x2
        self.y1 = self.y1 if y1 is None else y1
        self.y2 = self.y2 if y2 is None else y2
        self._update()

    def set_spacing(self, x_spacing=None, y_spacing=None):
        """set the fineness of the grid spacing and update the coordinate points"""
        self.x_spacing = self.x_spacing if x_spacing is None else x_spacing
        self.y_spacing = self.y_spacing if y_spacing is None else y_spacing
        self._update()

    def _update(self):
        """
        Update the number of points based on the spacing, and then the points
        """
        numx = int((self.x2 - self.x1) / self.x_spacing)
        numy = int((self.y2 - self.y1) / self.y_spacing)
        self.num_points = np.array([numx, numy])
        if self.offset:
            xpoints = np.linspace(
                self.x1 + (self.x_spacing / 2), self.x2 - (self.x_spacing / 2), numx
            )
            ypoints = np.linspace(
                self.y1 + (self.y_spacing / 2), self.y2 - (self.y_spacing / 2), numy
            )
        else:
            xpoints = np.linspace(self.x1, self.x2, numx)
            ypoints = np.linspace(self.y1, self.y2, numy)
        self.points = [xpoints, ypoints]
        self.xp, self.yp = self.points
        X, Y = [grid.reshape(-1) for grid in np.meshgrid(*self.points)]
        xy_coords = np.array([np.array((x0, y0)) for x0, y0 in zip(X, Y)])
        zs = np.ones(xy_coords.shape[0]) * self.height
        self.coords = np.stack([xy_coords.T[0], xy_coords.T[1], zs]).T

    def plot_plane(self, fig=None, vmin=None, vmax=None, title=None):
        """Plot the image of the radiation pattern"""
        if fig is None:
            fig, ax = plt.subplots()
        title = "" if title is None else title
        if self.values is not None:
            vmin = self.values.min() if vmin is None else vmin
            vmax = self.values.max() if vmax is None else vmax
            extent = [self.y1, self.y2, self.x1, self.x2]
            # ratio = (self.y2 - self.y1) / (self.x2 - self.x1)
            # if ratio < 1:
            # orientation, location = "horizontal", "top"
            # else:
            # orientation, location = "vertical", "right"
            img = ax.imshow(self.values.T, extent=extent, vmin=vmin, vmax=vmax)
            cbar = fig.colorbar(
                img,
                pad=0.03,  # orientation=orientation, location=location
            )
            ax.set_title(title)
            cbar.set_label(self.units, loc="center")
        return fig



from setuptools import setup, find_packages

with open("README.md", "r") as fh:
    long_description = fh.read()
    setup(
        name="guv-calcs",
        url="https://github.com/jvbelenky/guv-calcs",
        version="0.0.7",
        author="J. Vivian Belenky",
        author_email="j.vivian.belenky@outlook.com",
        description="A library for carrying out fluence and irradiance calculations for germicidal UV (GUV) applications.",
        long_description=long_description,
        long_description_content_type="text/markdown",
        classifiers=[
            "Programming Language :: Python :: 3",
            "Operating System :: OS Independent",
            "License :: OSI Approved :: MIT License",
        ],
        packages=find_packages('src'),
        package_dir={'': 'src'},
        zip_safe=True,
        python_requires=">=3.8",
        install_requires=[
            "numpy",
            "scipy",
            "matplotlib",
            "plotly",
            "photompy"
        ],
    )
